(*
Author:  Christian Sternagel <c.sternagel@gmail.com> (2011-2015)
Author:  René Thiemann <rene.thiemann@uibk.ac.at> (2011-2015)
Author:  Sarah Winkler <sarah.winkler@uibk.ac.at> (2013)
Author:  Akihisa Yamada <akihisa.yamada@uibk.ac.at> (2017)
License: LGPL (see file COPYING.LESSER)
*)
theory Proof_Checker
  imports
    Check_Termination
    Check_Nontermination
    Check_CRP
    Check_Conditional_CRP
    Check_Completion_Proof
    Check_Equational_Proof
    Check_Complexity
    Check_Quasi_Reductive
    Check_AC_Termination
    LTS.IA_Instance    
    XTC_Problem
begin

datatype (dead 'f, dead 'l, dead 'v) "proof" =
  TRS_Termination_Proof "('f, 'l, 'v) trs_termination_proof"
| Complexity_Proof "('f,'l,'v) complexity_proof"
| DP_Termination_Proof "('f, 'l, 'v) dp_termination_proof"
| DP_Nontermination_Proof "('f, 'l, 'v) dp_nontermination_proof"
| TRS_Nontermination_Proof "('f, 'l, 'v) trs_nontermination_proof"
| FP_Termination_Proof "('f,'l,'v) fptrs_termination_proof"
| Relative_TRS_Nontermination_Proof "('f, 'l, 'v) reltrs_nontermination_proof"
| TRS_Confluence_Proof "('f, 'l, 'v) cr_proof"
| TRS_Non_Confluence_Proof "('f, 'l, 'v, 'v) ncr_proof"
| TRS_Non_Commutation_Proof "('f, 'l, 'v, 'v) ncomm_proof"
| TRS_Commutation_Proof "('f,'l, 'v) comm_proof"
| Completion_Proof "('f, 'l, 'v) completion_proof"
| Ordered_Completion_Proof "(('f,'l) lab, 'v) ordered_completion_proof"
| Equational_Proof "('f, 'l, 'v) equational_proof"
| Equational_Disproof "('f, 'l, 'v) equational_disproof"
| Quasi_Reductive_Proof "('f, 'l, 'v) quasi_reductive_proof"
| Conditional_CR_Proof "('f, 'l, 'v) conditional_cr_proof"
| Conditional_Non_CR_Proof "('f, 'l, 'v, 'v) conditional_ncr_proof"
| Tree_Automata_Closed_Proof "string ta_relation"
| AC_Termination_Proof "('f,'l,'v) ac_termination_proof"
| LTS_Termination_Proof "(IA.sig, 'v, IA.ty, string, string, hints) termination_proof"
| LTS_Safety_Proof "(IA.sig, 'v, IA.ty, string, string, string, hints) safety_proof"
| Infeasibility_Proof "(('f, 'l) lab, 'v, ('f,'l)lab redtriple_impl) infeasibility_proof"
| Unknown_Proof "('f,'l,'v) unknown_proof"
| Unknown_Disproof "('f,'l,'v) neg_unknown_proof"

hide_const (open) eq (* eq from HOL-Algebra congruence is here used as variable name for an equation *)

definition "proof_to_string proof =
  (case proof of
    TRS_Termination_Proof _ \<Rightarrow> STR ''TRS termination proof''
  | Complexity_Proof _ \<Rightarrow> STR ''complexity proof''
  | DP_Termination_Proof _ \<Rightarrow> STR ''DP termination proof''
  | DP_Nontermination_Proof _ \<Rightarrow> STR ''DP nontermination proof''
  | TRS_Nontermination_Proof _ \<Rightarrow> STR ''TRS nontermination proof''
  | FP_Termination_Proof _ \<Rightarrow> STR ''FP TRS termination proof''
  | Relative_TRS_Nontermination_Proof _ \<Rightarrow> STR ''relative TRS nontermination proof''
  | TRS_Confluence_Proof _ \<Rightarrow> STR ''TRS confluence proof''
  | TRS_Non_Confluence_Proof _ \<Rightarrow> STR ''TRS nonconfluence proof''
  | TRS_Non_Commutation_Proof _ \<Rightarrow> STR ''TRS noncommutation proof''
  | TRS_Commutation_Proof _ \<Rightarrow> STR ''TRS commutation proof''
  | Completion_Proof _ \<Rightarrow> STR ''completion proof''
  | Ordered_Completion_Proof _ \<Rightarrow> STR ''ordered completion proof''
  | Equational_Proof _ \<Rightarrow> STR ''equational proof''
  | Equational_Disproof _ \<Rightarrow> STR ''equational disproof''
  | Quasi_Reductive_Proof _ \<Rightarrow> STR ''quasi-reductive proof''
  | Conditional_CR_Proof _ \<Rightarrow> STR ''conditional confluence proof''
  | Conditional_Non_CR_Proof _ \<Rightarrow> STR ''conditional nonconfluence proof''
  | Tree_Automata_Closed_Proof _ \<Rightarrow> STR ''tree automata closed proof''
  | AC_Termination_Proof _ \<Rightarrow> STR ''AC termination proof''
  | LTS_Termination_Proof _ \<Rightarrow> STR ''LTS termination proof''
  | LTS_Safety_Proof _ \<Rightarrow> STR ''LTS safety proof''
  | Infeasibility_Proof _ \<Rightarrow> STR ''Infeasibility proof''
  | Unknown_Proof _ \<Rightarrow> STR ''unknown proof''
  | Unknown_Disproof _ \<Rightarrow> STR ''unknown disproof'')"


primrec cert_assms :: "bool \<Rightarrow> ('f, 'l, 'v) proof \<Rightarrow> (('f, 'l, 'v) assm) list" where
  "cert_assms a (TRS_Termination_Proof p) = sn_assms a p"
| "cert_assms a (FP_Termination_Proof p) = sn_fp_assms a p"
| "cert_assms a (TRS_Confluence_Proof p) = cr_assms a p"
| "cert_assms a (TRS_Non_Confluence_Proof p) = ncr_assms a p"
| "cert_assms a (Completion_Proof p) = completion_assms a p"
| "cert_assms a (Ordered_Completion_Proof p) = []"
| "cert_assms a (DP_Termination_Proof p) = dp_assms a p"
| "cert_assms a (Equational_Proof p) = equational_assms a p"
| "cert_assms a (Equational_Disproof p) = equational_dis_assms a p"
| "cert_assms a (Quasi_Reductive_Proof p) = quasi_reductive_assms a p"
| "cert_assms a (Conditional_CR_Proof p) = conditional_cr_assms a p"
| "cert_assms a (Conditional_Non_CR_Proof p) = conditional_ncr_assms a p"
| "cert_assms a (Unknown_Proof p) = unknown_assms a p"
| "cert_assms a (Unknown_Disproof p) = neg_unknown_assms a p"
| "cert_assms a (Relative_TRS_Nontermination_Proof p) = neg_rel_sn_assms a p"
| "cert_assms a (TRS_Nontermination_Proof p) = neg_sn_assms a p"
| "cert_assms a (DP_Nontermination_Proof p) = neg_dp_assms a p"
| "cert_assms a (Complexity_Proof p) =  complexity_assms a p"
| "cert_assms a (LTS_Termination_Proof _) = []"
| "cert_assms a (LTS_Safety_Proof _) = []"
| "cert_assms a (Infeasibility_Proof _) = []"
| "cert_assms a (Tree_Automata_Closed_Proof _) =  []"
| "cert_assms a (AC_Termination_Proof _) =  []"
| "cert_assms a (TRS_Non_Commutation_Proof p) = []"
| "cert_assms a (TRS_Commutation_Proof p) = comm_assms a p"


lemma cert_assms_False[simp]: "cert_assms False p = []" by (cases p) simp_all

datatype cert_result = Certified | Unsupported String.literal | Error String.literal

context fixes
  I :: "('tp, ('f::{showl, compare_order,countable}, label_type) lab, string) tp_ops" and
  J :: "('dpp, ('f, label_type) lab, string) dpp_ops" and
  K :: "('ac_tp, ('f, label_type) lab, string) ac_tp_ops" and
  L :: "('ac_dpp, ('f, label_type) lab, string) ac_dpp_ops"
begin


definition check_cert
  :: "bool \<Rightarrow> (('f, label_type) lab, string) input \<Rightarrow> (('f, label_type) lab, string) claim \<Rightarrow> ('f, label_type, string) proof \<Rightarrow> showsl check"
where "check_cert a input claim proof \<equiv>
  let mismatch =
    (showsl_lit (STR ''Claiming '' + claim_to_string claim + STR '' of '' + input_to_string input + STR '' by '' + proof_to_string proof))
  in
  case input of Inn_TRS_input Q R S start \<Rightarrow> (
    case proof of TRS_Termination_Proof prf \<Rightarrow>
      if claim = Terminating \<or> claim = Anything then
        check_trs_termination_proof I J a (showsl_lit (STR ''1'')) (tp_ops.mk I default_nfs_trs (strategy_to_Q Q R) R S) prf
      else error mismatch
    | Complexity_Proof prf \<Rightarrow> (
      case claim of Upperbound ub \<Rightarrow>
        let cm = translate_complexity_claim R S start in
        let cc = Comp_Poly ub in
        check_complexity_proof I a (showsl_lit (STR ''1'')) (tp_ops.mk I True (strategy_to_Q Q (R @ S)) R S, cm, cc) prf
      | _ \<Rightarrow> error mismatch
      )
    | TRS_Nontermination_Proof prf \<Rightarrow>
      if claim = Nonterminating \<or> claim = Anything then
        check (start = start_term.Full) mismatch \<then>
        check (S = []) mismatch \<then>
        check_trs_nontermination_proof I J a (showsl_lit (STR ''1'')) (tp_ops.mk I default_nfs_nt_trs (strategy_to_Q Q R) R []) prf
      else error mismatch
    | Relative_TRS_Nontermination_Proof prf \<Rightarrow>
      if claim = Nonterminating \<or> claim = Anything then
        check (start = start_term.Full) mismatch \<then>
        check_reltrs_nontermination_proof I J a (showsl_lit (STR ''1'')) (tp_ops.mk I default_nfs_nt_trs (strategy_to_Q Q R) R S) prf
      else error mismatch
    | TRS_Confluence_Proof prf \<Rightarrow>
      if claim = Confluent \<or> claim = Anything then
        check (start = start_term.Full) (showsl_lit (STR ''Confluence with start term not supported'')) \<then>
        check (Q = No_Strategy) (showsl_lit (STR ''Confluence under strategy not supported'')) \<then>
        check (S = []) (showsl_lit (STR ''Relative confluence not supported'')) \<then>
        check_cr_proof a (showsl_lit (STR ''1'')) I J R prf
      else error mismatch
    | TRS_Non_Confluence_Proof prf \<Rightarrow>
      if claim = Nonconfluent \<or> claim = Anything then
        check (start = start_term.Full) (showsl_lit (STR ''Nonconfluence with start term not supported'')) \<then>
        check (Q = No_Strategy) (showsl_lit (STR ''Confluence under strategy not supported'')) \<then>
        check (S = []) (showsl_lit (STR ''Relative nonconfluence not supported'')) \<then>
        check_ncr_proof a (showsl_lit (STR ''1'')) I J R prf
      else error mismatch
    | _ \<Rightarrow> error mismatch
  )
  | Commutation_input F R S \<Rightarrow> (case proof of
     TRS_Non_Commutation_Proof prf \<Rightarrow> 
      if claim = Noncommuting \<or> claim = Anything then check_ncomm_proof (showsl_lit (STR ''1'')) F R S prf
      else error mismatch
     | TRS_Commutation_Proof prf \<Rightarrow> 
      if claim = Commuting \<or> claim = Anything then check_comm_proof I J a (showsl_lit (STR ''1'')) R S prf
      else error mismatch
     | _ \<Rightarrow> error mismatch)
  | FP_TRS_input strat R \<Rightarrow> (
    case proof of FP_Termination_Proof prf \<Rightarrow>
      if claim = Terminating \<or> claim = Anything then
        check_fptrs_termination_proof I J a (showsl_lit (STR ''1'')) (fp_strategy_to_fp_impl strat R, R) prf
      else error mismatch
    | TRS_Nontermination_Proof prf \<Rightarrow>
      if claim = Nonterminating \<or> claim = Anything then
        check_fp_nontermination_proof I J a (showsl_lit (STR ''1'')) (fp_strategy_to_fp_impl strat R, R) prf
      else error mismatch
    | _ \<Rightarrow> error mismatch
  )
  | CPX_input Q R S cm cc \<Rightarrow> ( \<comment> \<open>TODO: improve CPF and remove\<close>
    case proof of Complexity_Proof prf \<Rightarrow>
      ( case claim of Upperbound ub \<Rightarrow>
          check (cc = Comp_Poly ub) (showsl_lit (STR ''complexity class mismatch''))
        | Anything \<Rightarrow> Inr ()
        | _ \<Rightarrow> error mismatch
      ) \<then>
      check_complexity_proof I a (showsl_lit (STR ''1'')) (tp_ops.mk I True (strategy_to_Q Q (R @ S)) R S, cm, cc) prf
    | _ \<Rightarrow> error mismatch
  )
  | DP_input m P Q R \<Rightarrow> (
    case proof of (DP_Termination_Proof prf) \<Rightarrow>
      if claim = Terminating \<or> claim = Anything then
        check_dp_termination_proof I J a (showsl_lit (STR ''1'')) (dpp_ops.mk J default_nfs_dp m P [] (strategy_to_Q Q R) [] R) prf
      else error mismatch
    | DP_Nontermination_Proof prf \<Rightarrow>
      if claim = Nonterminating \<or> claim = Anything then
        check_dp_nontermination_proof I J a (showsl_lit (STR ''1'')) (dpp_ops.mk J default_nfs_nt_dp m P [] (strategy_to_Q Q R) [] R) prf
      else error mismatch
    | _ \<Rightarrow> error mismatch
  )
  | COMP_input E R \<Rightarrow> (
    case proof of Completion_Proof prf \<Rightarrow>
      if claim = Completed \<or> claim = Anything then
        check_completion_proof a (showsl_lit (STR ''1'')) I J E R prf
      else error mismatch
    | _ \<Rightarrow> error mismatch
  )
  | OCOMP_input E\<^sub>0 E R ord \<Rightarrow> (
    case proof of Ordered_Completion_Proof prf \<Rightarrow>
      if claim = Ordered_Completed \<or> claim = Anything then
        check_ordered_completion_proof_ext (showsl_lit (STR ''1'')) E\<^sub>0 E R ord prf
      else error mismatch
    | _ \<Rightarrow> error mismatch
  )
  | EQ_input E eq \<Rightarrow> (
    case proof of Equational_Proof prf \<Rightarrow>
      if claim = Yes \<or> claim = Anything then
        check_equational_proof a (showsl_lit (STR ''1'')) I J E eq prf
      else error mismatch
    | Equational_Disproof prf \<Rightarrow>
      if claim = No \<or> claim = Anything then
        check_equational_disproof a (showsl_lit (STR ''1'')) I J E eq prf
      else error mismatch
    | _ \<Rightarrow> error mismatch
  )
  | CTRS_input ctrs \<Rightarrow> (
    case proof of Quasi_Reductive_Proof prf \<Rightarrow>
      if claim = Terminating \<or> claim = Anything then
        check_quasi_reductive_proof a (showsl_lit (STR ''1'')) I J ctrs prf
      else error mismatch
    | Conditional_CR_Proof prf \<Rightarrow>
      if claim = Confluent \<or> claim = Anything then
        check_conditional_cr_proof a (showsl_lit (STR ''1'')) I J ctrs prf
      else error mismatch
    | Conditional_Non_CR_Proof prf \<Rightarrow>
      if claim = Nonconfluent \<or> claim = Anything then
        check_conditional_ncr_proof a (showsl_lit (STR ''1'')) I J ctrs prf
      else error mismatch
    | _ \<Rightarrow> error mismatch
  )
  | TA_input ta R \<Rightarrow> (
    case proof of Tree_Automata_Closed_Proof prf \<Rightarrow>
      if claim = Yes \<or> claim = Anything then
        tree_aut_trs_closed ta prf R
      else error mismatch
    | _ \<Rightarrow> error mismatch
  )
  | AC_input R A C \<Rightarrow> (
    case proof of AC_Termination_Proof prf \<Rightarrow>
      if claim = Terminating \<or> claim = Anything then
        check_ac_termination_proof L K (showsl_lit (STR ''1'')) (ac_tp_ops.mk K R A C) prf
      else error mismatch
    | _ \<Rightarrow> error mismatch
  )
  | LTS_input r \<Rightarrow> (
    case proof of LTS_Termination_Proof prf \<Rightarrow>
      if claim = Terminating \<or> claim = Anything then
        IA.check_termination r prf
      else error mismatch
    | _ \<Rightarrow> error mismatch
  )
  | LTS_safety_input r e \<Rightarrow> (
    case proof of LTS_Safety_Proof prf \<Rightarrow>
      if claim = Yes \<or> claim = Anything then
        IA.check_safety r e prf
      else error mismatch
    | _ \<Rightarrow> error mismatch
  )
  | Infeasibility_input c r \<Rightarrow> (
    case proof of Infeasibility_Proof prf \<Rightarrow>
      if claim = Yes \<or> claim = Anything then
        Infeasibility.check_infeasible' c r prf
      else error mismatch
    | _ \<Rightarrow> error mismatch
  )
  | Unknown_input ur \<Rightarrow> (
    case proof of Unknown_Proof prf \<Rightarrow>
      check_unknown_proof I J a (showsl_lit (STR ''1'')) ur prf
    | Unknown_Disproof prf \<Rightarrow>
      check_unknown_disproof I J a (showsl_lit (STR ''1'')) ur prf
    | _ \<Rightarrow> error mismatch
  )"

definition "certify_cert_problem a input claim proof \<equiv>
  case check_cert a input claim proof of
    Inl err \<Rightarrow> Error (err (STR ''''))
  | _ \<Rightarrow> Certified"

context
  assumes I: "tp_spec I" and J: "dpp_spec J" and K: "ac_tp_spec K" and L: "ac_dpp_spec L"
begin

lemma check_cert_with_assms_sound:
  assumes answer: "isOK (check_cert a input claim cp)"
      and a: "\<forall>p\<in>set (cert_assms a cp). holds p"
  shows "desired_property input claim"
proof-
  let ?i = "showsl_lit (STR ''1'')"
  note [simp] = check_cert_def desired_property_def fp_strategy_to_fp_impl_def
  interpret tp_spec I by (rule I)
  show ?thesis
  proof (cases input)
    case [simp]: (Inn_TRS_input Q R S start)
    show ?thesis
    proof (cases cp)
      case [simp]: (TRS_Termination_Proof prof)
      let ?tp = "tp_ops.mk I default_nfs_trs (strategy_to_Q Q R) R"
      from answer have ok: "isOK (check_trs_termination_proof I J a ?i (?tp S) prof)" by auto
      from answer check_trs_termination_proof_with_assms[OF I J ok] a
      show ?thesis by (auto split:claim.split)
    next
      case [simp]: (Complexity_Proof prof)
      let ?Q = "strategy_to_Q Q (R @ S)"
      let ?tp = "tp_ops.mk I True ?Q R S"
      show ?thesis
      proof (cases claim)
        case [simp]: (Upperbound ub)
        from answer have ok: "isOK (check_complexity_proof I a ?i (?tp,translate_complexity_claim R S start,Comp_Poly ub) prof)" 
          by auto
        from answer check_complexity_proof_with_assms[OF I infinite_lab ok] a
        show ?thesis by auto
      qed (insert answer, auto)
    next
      case [simp]: (TRS_Nontermination_Proof "prf")
      let ?tp = "tp_ops.mk I default_nfs_nt_trs (strategy_to_Q Q R) R []"
      from answer have ok: "isOK (check_trs_nontermination_proof I J a ?i ?tp prf)" and [simp]: "S = []" by auto
      from answer check_trs_nontermination_proof_with_assms[OF I J _ ok] a
      show ?thesis unfolding mk_simps by (auto split:claim.split)
    next
      case [simp]: (Relative_TRS_Nontermination_Proof "prf")
      let ?tp = "tp_ops.mk I default_nfs_nt_trs (strategy_to_Q Q R) R S"
      from answer have ok: "isOK (check_reltrs_nontermination_proof I J a ?i ?tp prf)"
        unfolding Relative_TRS_Nontermination_Proof
        by (cases "check_reltrs_nontermination_proof I J a ?i ?tp prf") auto
      from answer check_reltrs_nontermination_proof_with_assms[OF I J _ ok] a
        ok show ?thesis unfolding mk_simps by (auto split:claim.split)
    next
      case [simp]: (TRS_Confluence_Proof "prf")
      from answer have ok: "isOK (check_cr_proof a ?i I J R prf)"
       and [simp]: "Q = No_Strategy" "S = []" "start = start_term.Full" by auto
      from answer check_cr_proof_with_assms_sound[OF I J _ ok] a
      show ?thesis by (auto split:claim.split)
    next
      case [simp]: (TRS_Non_Confluence_Proof "prf")
      from answer have ok: "isOK (check_ncr_proof a ?i I J R prf)"
       and [simp]: "Q = No_Strategy" "S = []" "start = start_term.Full" by auto
      from answer check_ncr_proof_with_assms_sound[OF I J _ ok] a
      show ?thesis by (auto split:claim.split)
    qed (insert answer, auto)
  next
    case [simp]: (FP_TRS_input strat R)
    show ?thesis
    proof (cases cp)
      case [simp]: (FP_Termination_Proof "prf")
      let ?tp = "(fp_strategy_to_fp_impl strat R,R)"
      from answer have ok: "isOK (check_fptrs_termination_proof I J a ?i ?tp prf)" by auto
      from answer check_fptrs_termination_proof_with_assms[OF I J ok] a
      show ?thesis 
        by (auto simp: ostep_iff_fpstep_impl[of "map fst R",simplified] SN_fpstep_def csstep_iff_fpstep_impl split: fp_strategy.split claim.split)
    next
      case [simp]: (TRS_Nontermination_Proof "prf")
      have id: "lhss (set R) = set (map fst R)" by auto
      let ?p = "fp_strategy_to_fp_impl strat R"
      from answer have ok: "isOK (check_fp_nontermination_proof I J a ?i (?p,R) prf)" by auto
      from answer check_fp_nontermination_proof_with_assms[OF I J _ ok] a
      show ?thesis
        by (auto simp: ostep_iff_fpstep_impl[of "map fst R",simplified] SN_fpstep_def csstep_iff_fpstep_impl split: fp_strategy.split claim.split)
    qed (insert answer, auto)
  next
    case [simp]: (DP_input m P Q R)
    show ?thesis
    proof (cases cp)
      case [simp]: (DP_Termination_Proof "prf")
      let ?dp = "dpp_ops.mk J default_nfs_dp m P [] (strategy_to_Q Q R) [] R"
      from answer have ok: "isOK (check_dp_termination_proof I J a ?i ?dp prf)" by auto
      interpret J: dpp_spec J by (rule J)
      from answer check_dp_termination_proof_with_assms[OF I J ok] a
      show ?thesis by auto
    next
      case [simp]: (DP_Nontermination_Proof "prf")
      let ?dp = "dpp_ops.mk J default_nfs_nt_dp m P [] (strategy_to_Q Q R) [] R"
      from answer have ok: "isOK (check_dp_nontermination_proof I J a ?i ?dp prf)" by auto
      interpret J: dpp_spec J by (rule J)
      from answer check_dp_nontermination_proof_with_assms[OF I J _ ok] a
      show ?thesis by (auto simp del: infinite_dpp.simps)
    qed (insert answer, auto)
  next
    case [simp]: (COMP_input E R)
    show ?thesis
    proof (cases cp)
      case [simp]: (Completion_Proof "prf")
      from answer have ok: "isOK (check_completion_proof a ?i I J E R prf)" by auto
      from answer check_completion_proof_with_assms_sound[OF I J _ ok] a
      show ?thesis by auto
    qed (insert answer, auto)
  next
    case [simp]: (OCOMP_input E\<^sub>0 E R ro)
    show ?thesis
    proof (cases cp)
      case [simp]: (Ordered_Completion_Proof "prf")
      from answer have "isOK (check_ordered_completion_proof_ext ?i E\<^sub>0 E R ro prf)" by auto
      with answer check_ordered_completion_proof_ext_sound[OF this] a show ?thesis by auto
    qed (insert answer, auto)
  next
    case [simp]: (EQ_input E eq)
    show ?thesis
    proof (cases cp)
      case [simp]: (Equational_Proof "prf")
      from answer have ok: "isOK (check_equational_proof a ?i I J E eq prf)" by auto
      from answer check_equational_proof_with_assms_sound[OF I J _ ok] a
      show ?thesis by auto
    next
      case [simp]: (Equational_Disproof "prf")
      from answer have ok: "isOK (check_equational_disproof a ?i I J E eq prf)" by auto
      from answer check_equational_disproof_with_assms_sound[OF I J _ ok] a
      show ?thesis by auto
    qed (insert answer, auto)
  next
    case [simp]: (CTRS_input ctrs)
    show ?thesis
    proof (cases cp)
      case [simp]: (Quasi_Reductive_Proof "prf")
      from answer have ok: "isOK (check_quasi_reductive_proof a ?i I J ctrs prf)" by auto
      from answer check_quasi_reductive_proof_with_assms_sound[OF I J _ ok] quasi_reductive_SN a
      show ?thesis by auto
    next
      case [simp]: (Conditional_CR_Proof "prf")
        from answer have ok: "isOK (check_conditional_cr_proof a ?i I J ctrs prf)" by auto
        from answer check_conditional_cr_proof_with_assms_sound[OF I J _ ok] a
        show ?thesis by auto
      next
        case [simp]: (Conditional_Non_CR_Proof "prf")
      from answer have ok: "isOK (check_conditional_ncr_proof a ?i I J ctrs prf)" by auto
      from answer check_conditional_ncr_proof_with_assms_sound[OF I J _ ok] a
      show ?thesis by auto
    qed (insert answer, auto)
  next
    case [simp]: (TA_input ta r)
    show ?thesis
    proof (cases cp)
      case [simp]: (Tree_Automata_Closed_Proof "prf")
      from answer have ok: "isOK (tree_aut_trs_closed ta prf r)" by auto
      from answer tree_aut_trs_closed[OF ok]
      show ?thesis by (cases claim, auto)
    qed (insert answer, auto)
  next
    case [simp]: (AC_input r a c)
    show ?thesis
    proof (cases cp)
      case [simp]: (AC_Termination_Proof "prf")
      let ?tp = "ac_tp_ops.mk K r a c"
      interpret K: ac_tp_spec K by fact
      from answer have ok: "isOK (check_ac_termination_proof L K ?i ?tp prf)" by auto
      from answer check_ac_termination_proof[OF L K this]
      show ?thesis by auto
    qed (insert answer, auto)
  next
    case [simp]: (LTS_input r)
    show ?thesis
    proof (cases cp)
      case [simp]: (LTS_Termination_Proof "prf")
      from answer have ok: "isOK (IA.check_termination r prf)" by auto
      from answer IA.check_termination[OF this]
      show ?thesis by (auto simp: lts_of_def)
    qed (insert answer, auto)
  next
    case [simp]: (LTS_safety_input r e)
    show ?thesis
    proof (cases cp)
      case [simp]: (LTS_Safety_Proof "prf")
      from answer have ok: "isOK (IA.check_safety r e prf)" by auto
      from answer IA.check_safety[OF this]
      show ?thesis by (auto simp: lts_of_def)
    qed (insert answer, auto)
  next
    case [simp]: (Infeasibility_input c r)
    show ?thesis
    proof (cases cp)
      case [simp]: (Infeasibility_Proof "prf")
      from answer Infeasibility.check_infeasible'[of c r "prf"]
      show ?thesis by (auto simp: conds_sat_conds_n_sat)
    qed (insert answer, auto)
  next
    case [simp]: (Commutation_input F R S)
    show ?thesis
    proof (cases cp)
      case [simp]: (TRS_Non_Commutation_Proof "prf")
      from answer obtain i where ok: "isOK (check_ncomm_proof i F R S prf)" by auto
      from answer check_ncomm_proof_sound[OF this] show ?thesis by auto
    next
      case [simp]: (TRS_Commutation_Proof "prf")
      from answer obtain i where ok: "isOK (check_comm_proof I J a i R S prf)" by auto
      from answer check_comm_proof_with_assms_sound[OF I J this] a show ?thesis by (auto intro!: commute_imp_sig_commute)
    qed (insert answer, auto)
  next
    case [simp]: (Unknown_input up)
    show ?thesis
    proof (cases cp)
      case [simp]: (Unknown_Proof "prf")
      from answer have ok: "isOK (check_unknown_proof I J a ?i up prf)" by auto
      from check_unknown_proof_with_assms[OF I J ok] a
      show ?thesis by auto
    next
      case [simp]: (Unknown_Disproof "prf")
      from answer have ok: "isOK (check_unknown_disproof I J a ?i up prf)" by auto
      from check_unknown_disproof_with_assms[OF I J ok] a
      show ?thesis by auto
    qed (insert answer, auto)
  next
    case [simp]: (CPX_input Q R S cm cc)
    show ?thesis
    proof (cases cp)
      case [simp]: (Complexity_Proof prof)
      show ?thesis
      proof (cases claim)
        case [simp]: (Upperbound ub)
        let ?Q = "strategy_to_Q Q (R @ S)"
        let ?tp = "tp_ops.mk I True ?Q R S"
        from answer have ok: "isOK (check_complexity_proof I a ?i (?tp,cm,cc) prof)"
         and [simp]: "cc = Comp_Poly ub"
          by auto
        from check_complexity_proof_with_assms[OF I infinite_lab ok] a
        show ?thesis by auto
      qed (insert answer, auto)
    qed (insert answer, auto)
  qed
qed

lemma certify_cert_problem_sound:
  assumes answer: "certify_cert_problem False input claim proof = Certified"
  shows "desired_property input claim"
  using answer
  by (cases "check_cert False input claim proof", auto simp: certify_cert_problem_def intro: check_cert_with_assms_sound)

end
end
end
